﻿// Created by Windward Studios - no copyright is claimed. This code can be used in
// any manner by anyone for any reason. There is no copyright of any kind on it. You may
// use it in commercial products. You may change it without sharing those changes.
// We ask that you keep the "created by Windward Studios" in a comment at the top.

using System;
using System.Collections;
using System.Data.Common;
using Kailua.net.windward.utils;

namespace net.windward.utils.ado.Access
{
	///<summary>
	/// A lightweight handler of functionality that the ODBC and OleDb drivers can use as well as the main implementations.
	///</summary>
	public class WrAccessSyntax : IWrSyntax
	{

		private readonly DbProviderFactory provider;
		private readonly string odbcProvider;
		private static readonly Hashtable reservedWords = new Hashtable();

		static WrAccessSyntax()
		{
			foreach (string str in _reservedWords)
				reservedWords.Add(str.ToUpper(), str.ToUpper());
		}

		/// <summary>
		/// Create the object.
		/// </summary>
		/// <param name="provider">The provider for this connector.</param>
		/// <param name="odbcProvider">The ODBC Provider used.</param>
		public WrAccessSyntax(DbProviderFactory provider, string odbcProvider)
		{
			this.provider = provider;
			this.odbcProvider = odbcProvider;
		}

		/// <summary>
		/// A connection string for a database on a server.
		/// </summary>
		/// <param name="server">The server to access (ex: localhost).</param>
		/// <param name="database">The database to connect to.</param>
		/// <param name="credentials">The user credentials to access the database.</param>
		/// <returns>The connection string for this database on this server.</returns>
		public string ConnectionString(string server, string database, WrCredentials credentials)
		{

			Trap.trap();
			DbConnectionStringBuilder connStrBuilder = provider.CreateConnectionStringBuilder();
			connStrBuilder.Add("Driver", odbcProvider);
			Trap.trap(!credentials.UseWindowsIdentity);
			connStrBuilder.Add("Dbq", FileUtils.FullPath(server));
			if (!credentials.UseWindowsIdentity)
			{
				Trap.trap();
				connStrBuilder.Add("Uid", credentials.Username);
				connStrBuilder.Add("Pwd", credentials.Password);
			}
			else Trap.trap();
			return connStrBuilder.ConnectionString;
		}

		/// <summary>
		/// Convert a select to normalize the "top 10" syntax
		/// </summary>
		/// <param name="select">The select with our standard usage</param>
		/// <returns>The select for this database.</returns>
		public string ConvertSelect(string select)
		{
			int pos = select.ToLower().LastIndexOf(" top ");
			if (pos == -1)
				return select;

			int posSelect = select.IndexOf("select ");
			return select.Substring(0, posSelect + 7).Trim() + " " + select.Substring(pos).Trim() + " " +
				select.Substring(posSelect + 7, pos - posSelect - 7).Trim();
		}

		/// <summary>
		/// Create a safe table name. Generally needed only if the name has a space in it or is a reserved name.
		/// </summary>
		/// <param name="tableName">The true table name.</param>
		/// <returns>The table name as it should be used in a select.</returns>
		public string SafeTableName(string tableName)
		{
			// schema.table name -> schema.[table name]
			int dot = tableName.IndexOf('.');
			if (dot == -1)
				return MakeNameSafe(tableName);
			Trap.trap();
			return tableName.Substring(0, dot) + '.' + MakeNameSafe(tableName.Substring(dot + 1));
		}

		/// <summary>
		/// Create a safe column name. Generally needed only if the name has a space in it or is a reserved name.
		/// </summary>
		/// <param name="columnName">The true column name.</param>
		/// <returns>The table name as it should be used in a select.</returns>
		public string SafeColumnName(string columnName)
		{
			return MakeNameSafe(columnName);
		}

		/// <summary>
		/// Create a safe table name. Generally needed only if the name has a space in it or is a reserved name.
		/// </summary>
		/// <param name="tableName">The true table name.</param>
		/// <param name="columnName">The true column name.</param>
		/// <returns>The table.column name as it should be used in a select.</returns>
		public string SafeTableColumnName(string tableName, string columnName)
		{
			return SafeTableName(tableName) + '.' + SafeColumnName(columnName);
		}

		// from http://office.microsoft.com/en-us/access/HA012315091033.aspx
		private static readonly string[] _reservedWords = { "ABSOLUTE", "ANY",
		                                                 	"ADD", "ARE",
		                                                 	"ADMINDB", "AS",
		                                                 	"ALL", "ASC",
		                                                 	"ALPHANUMERIC", "ASSERTION",
		                                                 	"ALTER", "AUTHORIZATION",
		                                                 	"AUTOINCREMENT",
		                                                 	"AND", "AVG",
		                                                 	"BEGIN", "COLLATION",
		                                                 	"BETWEEN", "COLUMN",
		                                                 	"BINARY", "COMMIT",
		                                                 	"BIT", "COMP", "COMPRESSION",
		                                                 	"BIT_LENGTH", "CONNECT",
		                                                 	"BOOLEAN", "CONNECTION",
		                                                 	"BOTH", "CONSTRAINT", "CONSTRAINTS",
		                                                 	"BY", "CONTAINER",
		                                                 	"BYTE", "CONTAINS",
		                                                 	"CASCADE", "CONVERT",
		                                                 	"CATALOG", "COUNT",
		                                                 	"CHAR", "CHARACTER", "COUNTER",
		                                                 	"CHAR_LENGTH", "CREATE",
		                                                 	"CHARACTER_LENGTH", "CURRENCY",
		                                                 	"CHECK", "CURRENT_DATE",
		                                                 	"CLOSE", "CURRENT_TIME",
		                                                 	"CLUSTERED", "CURRENT_TIMESTAMP",
		                                                 	"COALESCE", "CURRENT_USER",
		                                                 	"COLLATE", "CURSOR",
		                                                 	"DATABASE", "DISALLOW",
		                                                 	"DATE", "DISCONNECT",
		                                                 	"DATETIME", "DISTINCT",
		                                                 	"DAY", "DISTINCTROW",
		                                                 	"DEC", "DECIMAL", "DOMAIN",
		                                                 	"DECLARE", "DOUBLE",
		                                                 	"DELETE", "DROP",
		                                                 	"DESC",
		                                                 	"EQV", "FOREIGN",
		                                                 	"EXCLUSIVECONNECT", "FROM",
		                                                 	"EXEC", "EXECUTE", "CLAUSE",
		                                                 	"EXISTS", "GENERAL",
		                                                 	"EXTRACT", "GRANT",
		                                                 	"FALSE", "GROUP",
		                                                 	"FETCH", "GUID",
		                                                 	"FIRST", "HAVING",
		                                                 	"FLOAT", "FLOAT8", "HOUR",
		                                                 	"FLOAT4",
		                                                 	"IDENTITY", "INPUT",
		                                                 	"IEEEDOUBLE", "INSENSITIVE",
		                                                 	"IEEESINGLE", "INSERT",
		                                                 	"IGNORE", "INTO",
		                                                 	"IMAGE", "INT", "INTEGER", "INTEGER4",
		                                                 	"IMP", "INTEGER1",
		                                                 	"INTEGER2",
		                                                 	"IN", "INTERVAL",
		                                                 	"INDEX",
		                                                 	"INDEXCREATEDB", "IS",
		                                                 	"INNER", "ISOLATION",
		                                                 	"JOIN", "LONGTEXT",
		                                                 	"KEY", "LOWER",
		                                                 	"LANGUAGE", "MATCH",
		                                                 	"LAST", "MAX",
		                                                 	"LEFT", "MEMO",
		                                                 	"LEVEL", "MIN",
		                                                 	"LIKE", "MINUTE",
		                                                 	"LOGICAL", "LOGICAL1", "MOD",
		                                                 	"LONG", "MONEY",
		                                                 	"LONGBINARY", "MONTH",
		                                                 	"LONGCHAR",
		                                                 	"NATIONAL", "OUTER",
		                                                 	"NCHAR", "OUTPUT",
		                                                 	"NONCLUSTERED", "OWNERACCESS",
		                                                 	"NOT", "PAD",
		                                                 	"NTEXT", "PARAMETERS",
		                                                 	"NULL", "PARTIAL",
		                                                 	"NUMBER", "PASSWORD",
		                                                 	"NUMERIC", "PERCENT",
		                                                 	"NVARCHAR", "PIVOT",
		                                                 	"OCTET_LENGTH", "POSITION",
		                                                 	"OLEOBJECT", "PRECISION",
		                                                 	"ON", "PREPARE",
		                                                 	"OPEN", "PRIMARY",
		                                                 	"OPTION", "PRIVILEGES",
		                                                 	"OR", "PROC", "PROCEDURE",
		                                                 	"ORDER", "PUBLIC",
		                                                 	"REAL", "SMALLDATETIME",
		                                                 	"REFERENCES", "SMALLINT",
		                                                 	"RESTRICT", "SMALLMONEY",
		                                                 	"REVOKE", "SOME",
		                                                 	"RIGHT", "SPACE",
		                                                 	"ROLLBACK", "SQL",
		                                                 	"SCHEMA", "SQLCODE", "SQLERROR", "SQLSTATE",
		                                                 	"SECOND", "STDEV",
		                                                 	"SELECT", "STDEVP",
		                                                 	"SELECTSCHEMA", "STRING",
		                                                 	"SELECTSECURITY", "SUBSTRING",
		                                                 	"SET", "SUM",
		                                                 	"SHORT", "SYSNAME",
		                                                 	"SINGLE", "SYSTEM_USER",
		                                                 	"SIZE",
		                                                 	"TABLE", "UPDATEOWNER",
		                                                 	"TABLEID", "UPDATESECURITY",
		                                                 	"TEMPORARY", "UPPER",
		                                                 	"TEXT", "USAGE",
		                                                 	"TIME", "USER",
		                                                 	"TIMESTAMP", "USING",
		                                                 	"TIMEZONE_HOUR", "VALUE",
		                                                 	"TIMEZONE_MINUTE", "VALUES",
		                                                 	"TINYINT", "VAR",
		                                                 	"TO", "VARBINARY",
		                                                 	"TOP", "VARCHAR",
		                                                 	"TRAILING", "VARP",
		                                                 	"TRANSACTION", "VARYING",
		                                                 	"TRANSFORM", "VIEW",
		                                                 	"TRANSLATE", "WHEN",
		                                                 	"TRANSLATION", "WHENEVER",
		                                                 	"TRIM", "WHERE",
		                                                 	"TRUE", "WITH",
		                                                 	"UNION", "WORK",
		                                                 	"UNIQUE", "XOR",
		                                                 	"UNIQUEIDENTIFIER", "YEAR",
		                                                 	"UNKNOWN", "YESNO",
		                                                 	"UPDATE", "ZONE",
		                                                 	"UPDATEIDENTITY"};

		private static string MakeNameSafe(string columnName)
		{
			if (columnName[0] == '[' && columnName[columnName.Length - 1] == ']')
				return columnName;

			bool makeSafe = reservedWords.ContainsKey(columnName.ToUpper());
			if (!makeSafe)
				foreach (char ch in columnName)
					if ((!Char.IsLetterOrDigit(ch)) && (ch != '_'))
					{
						makeSafe = true;
						break;
					}
			if (makeSafe)
			{
				if (columnName[0] == '[')
					return columnName + ']';
				if (columnName[columnName.Length - 1] == ']')
					return '[' + columnName;
				return '[' + columnName + ']';
			}

			return columnName;
		}
	}
}