﻿// Created by Windward Studios - no copyright is claimed. This code can be used in
// any manner by anyone for any reason. There is no copyright of any kind on it. You may
// use it in commercial products. You may change it without sharing those changes.
// We ask that you keep the "created by Windward Studios" in a comment at the top.

using System;
using System.Collections;
using System.Data.Common;

using System.Data;
using System.Data.OleDb;

namespace net.windward.utils.ado.Excel
{
	/// <summary>
	/// An Excel database.
	/// </summary>
	public class WrExcelDatabase : IWrDatabase
	{
        private readonly DbProviderFactory provider;
		private readonly string odbcProvider;
        private readonly string server;
        private static readonly Hashtable types;
        private readonly IWrSyntax syntax;

		private WrTable[] tables;

        static WrExcelDatabase()
		{
            types = new Hashtable();
            types.Add("System.Double", DbType.Double);
            types.Add("System.String", DbType.String);
            types.Add("System.DateTime", DbType.DateTime);
        }

		/// <summary>
		/// Create an Excel database object.
		/// </summary>
		/// <param name="provider">The provider for this database.</param>
		/// <param name="odbcProvider">The ODBC Provider used.</param>
		/// <param name="filename">The filename of the database (.MDB file).</param>
        public WrExcelDatabase(DbProviderFactory provider, string odbcProvider, string filename)
		{
            this.provider = provider;
            this.odbcProvider = odbcProvider;
            this.server = filename;
			syntax = new WrExcelSyntax(provider, odbcProvider);
		}

        /// <summary>
        /// The name of the database.
        /// </summary>
        public string Name
        {
            get { return server; }
        }

        /// <summary>
        /// The default schema for all objects in this database. Returns null if there is no default.
        /// </summary>
        public string DefaultSchema
        {
            get
            {
                return null;
            }
        }

		/// <summary>
		/// All tables in this database. Must call LoadMetadata first.
		/// </summary>
		public WrTable[] Tables
		{
			get { return tables; }
		}

		/// <summary>
		/// All views in this database. Must call LoadMetadata first.
		/// </summary>
		public WrView[] Views
		{
			get { return null; }
		}

		/// <summary>
		/// Return all stored procedures in this database. Must call LoadMetadata first.
		/// </summary>
		public WrStoredProcedure[] StoredProcedures
		{
			get { return null; }
		}

		/// <summary>
		/// Loads the metadata for this database.
		/// </summary>
		/// <param name="credentials">The user credentials to access the database.</param>
		/// <param name="owner">Determines what metadata objects are read and if full metadata is retrieved.</param>
		/// <param name="status">Pass the status of loading the metadata.</param>
		public void LoadMetadata(WrCredentials credentials, WrVendor.ELEM_OWNER owner, WrDatabase.LoadStatus status)
		{
			LoadMetadata(ConnectionString("INFORMATION_SCHEMA", credentials), owner, status);
		}

		/// <summary>
		/// Loads the metadata for this database.
		/// </summary>
		/// <param name="connStr">The connection string to access the database.</param>
		/// <param name="owner">Determines what metadata objects are read and if full metadata is retrieved.</param>
		/// <param name="status">Pass the status of loading the metadata.</param>
		public void LoadMetadata(string connStr, WrVendor.ELEM_OWNER owner, WrDatabase.LoadStatus status)
		{
			tables = GetTables(connStr, owner);
		}

        /// <summary>
        /// Return all tables in this database.
        /// </summary>
        /// <param name="connStr">The connection string to access the database.</param>
        /// <param name="owner">Determines what tables are returned and if metadata is also retrieved.</param>
        /// <returns>all tables in this database.</returns>
        private static WrTable[] GetTables(string connStr, WrVendor.ELEM_OWNER owner)
        {

            if ((owner & WrVendor.ELEM_OWNER.USER) == 0)
                return new WrTable[0];
            try
            {
                using (OleDbConnection conn = new OleDbConnection(connStr))
                {
                    conn.Open();
                    ArrayList rtn = new ArrayList();
                    DataTable tables = conn.GetOleDbSchemaTable(OleDbSchemaGuid.Tables, new object[] { null, null, null, "TABLE" });
                    foreach (DataRow row in tables.Rows)
                    {
                    	string tableName = (string) row[2];
						WrColumn[] columns = GetColumns(tableName, connStr, owner);
                        rtn.Add(new WrTable(null, tableName, "", false, columns));
                    }
                    return (WrTable[])rtn.ToArray(typeof(WrTable));
                }
            }
            catch (Exception)
            {
                return new WrTable[0];
            }
        }

        /// <summary>
        /// Return all columns in a table or view.
        /// </summary>
		/// <param name="tableName">The name of the table or view.</param>
        /// <param name="connStr">The connection string to access the database.</param>
        /// <param name="owner">Determines what columns are returned and if metadata is also retrieved.</param>
        /// <returns>all columns in a table or view.</returns>
		private static WrColumn[] GetColumns(string tableName, string connStr, WrVendor.ELEM_OWNER owner)
        {
            try
            {
                ArrayList rtn = new ArrayList();
                using (OleDbConnection conn = new OleDbConnection(connStr))
                {
                    conn.Open();
                    DataSet myDataSet = new DataSet();
					OleDbCommand myAccessCommand = new OleDbCommand(String.Format("select * from [{0}]", tableName), conn);
                    OleDbDataAdapter myDataAdapter = new OleDbDataAdapter(myAccessCommand);
                    

                    myDataAdapter.Fill(myDataSet, "Categories");
                    DataColumnCollection drc = myDataSet.Tables["Categories"].Columns;
                    foreach (DataColumn dc in drc)
                    {
                        string str = dc.DataType.ToString();
                        DbType typ = (DbType)(types.Contains(str) ? types[str] : DbType.Object);
                        rtn.Add(new WrColumn(dc.ColumnName, "", typ));
                    }
                }
                return (WrColumn[])rtn.ToArray(typeof(WrColumn));
            }
            catch (Exception)
            {
                return new WrColumn[0];
            }
        }

        private string ConnectionString(string dbName, WrCredentials credentials)
        {
            return WrExcelVendor.ConnectionString(provider, odbcProvider, server, credentials, true);
        }
	}
}
