// Created by Windward Studios - no copyright is claimed. This code can be used in
// any manner by anyone for any reason. There is no copyright of any kind on it. You may
// use it in commercial products. You may change it without sharing those changes.
// We ask that you keep the "created by Windward Studios" in a comment at the top.

using System.Collections;
using System.Data;
using System.Data.Common;
using Kailua.net.windward.utils;

namespace net.windward.utils.ado.MySql
{
	/// <summary>
	/// An instance of a MySql server running on a computer.
	/// </summary>
	public class WrMySqlServer : IWrServer
	{
		private readonly DbProviderFactory provider;
		private readonly string name;

		/// <summary>
		/// Create a MySql server object.
		/// </summary>
		/// <param name="provider">The provider for this server.</param>
		/// <param name="name">The name of this server.</param>
		public WrMySqlServer(DbProviderFactory provider, string name)
		{
			this.provider = provider;
			this.name = name;
		}

		#region IWrServer Members

		/// <summary>
		/// The name of the server.
		/// </summary>
		public string Name
		{
			get { return name; }
		}

		/// <summary>
		/// All databases on this server. If a provider does not have the concept of databases (Oracle)
		/// it will always return 1 database.
		/// </summary>
		/// <param name="credentials">The user credentials to access the database.</param>
		/// <returns>All databases on this server.</returns>
		public IWrDatabase[] GetDatabases(WrCredentials credentials)
		{
			ArrayList rtn = new ArrayList();
			using (IDbConnection sc = provider.CreateConnection())
			{
				DbConnectionStringBuilder connStrBuilder = provider.CreateConnectionStringBuilder();
				connStrBuilder.Add("server", name);
				connStrBuilder.Add("database", "INFORMATION_SCHEMA");
				connStrBuilder.Add("user id", credentials.Username);
				connStrBuilder.Add("password", credentials.Password);
				sc.ConnectionString = connStrBuilder.ConnectionString;
				sc.Open();
				using (IDbCommand cmd = sc.CreateCommand())
				{
					cmd.CommandType = CommandType.Text;
					cmd.CommandText = "show databases";

					using (IDataReader myReader = cmd.ExecuteReader())
					{
						while (myReader.Read())
							rtn.Add(new WrMySqlDatabase(provider, name, myReader.GetString(0)));
					}
				}
			}
			return (WrMySqlDatabase[]) rtn.ToArray(typeof (WrMySqlDatabase));
		}

		/// <summary>
		/// All databases on this server. If a provider does not have the concept of databases (Oracle)
		/// it will always return 1 database.
		/// </summary>
		/// <param name="connStr">The connection string to access the database.</param>
		/// <returns>All databases on this server.</returns>
		public IWrDatabase[] GetDatabases(string connStr)
		{

			Trap.trap();
			ArrayList rtn = new ArrayList();
			using (IDbConnection sc = provider.CreateConnection())
			{
				sc.ConnectionString = connStr;
				sc.Open();
				using (IDbCommand cmd = sc.CreateCommand())
				{
					cmd.CommandType = CommandType.Text;
					cmd.CommandText = "show databases";

					using (IDataReader myReader = cmd.ExecuteReader())
					{
						while (myReader.Read())
							rtn.Add(new WrMySqlDatabase(provider, name, myReader.GetString(0)));
					}
				}
			}
			return (WrMySqlDatabase[])rtn.ToArray(typeof(WrMySqlDatabase));
		}

		#endregion

		/// <summary>
		/// Create a database object.
		/// </summary>
		/// <param name="database">The name of the database.</param>
		public IWrDatabase CreateDatabase(string database)
		{
			return new WrMySqlDatabase(provider, name, database);
		}

		/// <summary>
		/// Compares the current instance with another object of the same type.
		/// </summary>
		/// <param name="obj">An object to compare with this instance. </param>
		/// <returns>a 32-bit signed integer that indicates the relative order of the objects being compared.</returns>
		public int CompareTo(object obj)
		{
			WrMySqlServer db = obj as WrMySqlServer;
			if (db == null)
				return 1;
			return name.CompareTo(db.name);
		}
	}
}
